/*
 * GkSu -- Gtk+ Frontend to su
 * Copyright (C) 2002 Gustavo Noronha Silva
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <pty.h>

#include <gtk/gtk.h>

#include "../config.h"

#include "defines.h"

#include "gksu.h"
#include "xauth.h"
#include "su.h"
#include "gui.h"
#include "util.h"

/**
 * su_do:
 * @conf: a #GkConf struct holding all the settings for
 * this run
 *
 * This could be considered one of the main functions in GKSu.
 * it is responsible for doing the 'user changing' magic by
 * calling ask_password() if it needs the target user's password
 *
 * Returns: 1 if failed, 0 if succeeded.
 */
int
su_do (GkConf *conf)
{
  int i = 0;
  gboolean auth_failed = FALSE;

  int fdpty;
  pid_t pid;

  gchar *password = NULL;
  gchar *pass = NULL;

  pid = forkpty (&fdpty, NULL, NULL, NULL);
  if (pid == 0)
    {
      gchar **cmd = g_malloc (sizeof(gchar*)*7);

      cmd[i] = g_strdup ("/bin/su"); i++;
      if (conf->login_shell)
	{
	  cmd[i] = g_strdup ("-"); i++;
	}
      cmd[i] = g_strdup (conf->user); i++;
      if (conf->keep_env)
	{
	  cmd[i] = g_strdup ("-p"); i++;
	}
      cmd[i] = g_strdup ("-c"); i++;
      /* needs to get X authorization prior to running the program */
      if (conf->debug)
	{
	  cmd[i] = 
	    g_strdup_printf ("/usr/bin/env -u XAUTHORITY=%s /usr/X11R6/bin/xauth add %s . \"%s\"; /usr/bin/env -u XAUTHORITY=%s %s",
			     conf->xauth_file, conf->display, 
			     conf->xauth, conf->xauth_file,
			     conf->command); i++;
	}
      else
	{
	  cmd[i] = 
	    g_strdup_printf ("/usr/bin/env -u XAUTHORITY=%s /usr/X11R6/bin/xauth add %s . \"%s\" >& /dev/null; /usr/bin/env -u XAUTHORITY=%s %s", 
			     conf->xauth_file, conf->display, 
			     conf->xauth, conf->xauth_file,
			     conf->command); i++;
	}
      cmd[i] = NULL;

      /* executes the command */
      if (execv (cmd[0], cmd) == -1)
	{
	  gk_dialog (GTK_MESSAGE_ERROR, 
		     _("Unable to run /bin/su: %s"),
		     strerror(errno));
	}

      for (i = 0 ; cmd[i] != NULL ; i++)
	g_free (cmd[i]);
      g_free(cmd);
    }
  else if (pid == -1)
    {
      gk_dialog (GTK_MESSAGE_ERROR, 
		 strerror(errno));
      return 1;
    }
  else
    {
      fd_set rfds;

      struct timeval tv;

      struct passwd *pwd = NULL;
      gint target_uid = -1;
      gint my_uid = 0;

      gchar buf[256];
      gint status;

      read (fdpty, buf, 256);
      
      my_uid = getuid();
      pwd = getpwnam (conf->user);
      if (pwd)
	target_uid = pwd->pw_uid;

      /* no need to ask for password if we're already root */
      if (my_uid != target_uid && my_uid)
	{
	  /* ask for password */
	  if (conf->message == NULL)
	    password = ask_password (conf, 
				     _("<b>I need %s's password to run:</b>\n"
				       "<i>%s</i>"), 
				     conf->user, conf->command);
	  else
	    password = ask_password (conf, conf->message);

	  if (password == NULL)
	    return 1;

	  /* we cannot pass 'password' as it has no \n */
	  pass = g_strdup_printf ("%s\n", password);
	  write (fdpty, pass, strlen(pass));

	  /* cleans the memory */
	  free_pass (pass);
	  free_pass (password);
	} 

      FD_ZERO(&rfds);
      FD_SET(fdpty, &rfds);

      while (!waitpid (pid, &status, WNOHANG))
	{
	  tv.tv_sec = 0;
	  tv.tv_usec = 100;
	  if(select (fdpty+1, &rfds, NULL, NULL, &tv) < 0)
	    {
	      gk_dialog (GTK_MESSAGE_ERROR, 
			 strerror(errno));
	      return 1;
	    }
	  if(FD_ISSET(fdpty, &rfds))
	    {
	      bzero(buf, 256);
	      read (fdpty, buf, 255);
	      fprintf (stderr, "%s", buf);
	    }
	  else
	    FD_SET(fdpty, &rfds);
	}

      while (1)
	{
	  if (!strncmp (buf, "su: Authentication failure", 26))
	    auth_failed = TRUE;
	  bzero(buf, 256);
	  if(read (fdpty, buf, 255) <= 0)
	    break;
	  fprintf (stderr, "%s", buf);
	}
      clean_dir (conf->dir);

      if (WIFEXITED(status))
	{
	  if (WEXITSTATUS(status))
	    {
	      if (auth_failed)
		gk_dialog (GTK_MESSAGE_ERROR,
			   _("Wrong password."));
	      else
		gk_dialog (GTK_MESSAGE_INFO,
			   _("Child terminated with %d status"),
			   WEXITSTATUS(status));
	      return 1;
	    }
	}
    }

  return 0;
}
